@@ -32,7 +32,7 @@
 #include <osmocom/gprs/gprs_bssgp.h>
 
 #include <osmocom/sgsn/debug.h>
-#include <osmocom/sgsn/gprs_gb.h>
+#include <osmocom/sgsn/gprs_ns.h>
 #include <osmocom/sgsn/gprs_llc.h>
 #include <osmocom/sgsn/sgsn.h>
 #include <osmocom/sgsn/gprs_sndcp.h>
@@ -41,6 +41,9 @@
 #include <osmocom/sgsn/gprs_sndcp_pcomp.h>
 #include <osmocom/sgsn/gprs_sndcp_dcomp.h>
 #include <osmocom/sgsn/gprs_sndcp_comp.h>
+#include <osmocom/sgsn/gprs_gmm.h>
+#include <osmocom/sgsn/mmctx.h>
+#include <osmocom/sgsn/gtp.h>
 
 #define DEBUG_IP_PACKETS 0	/* 0=Disabled, 1=Enabled */
 
@@ -66,7 +69,7 @@ static uint16_t calc_ip_csum(uint8_t *data, int len)
 }
 
 /* Calculate TCP/IP checksum */
-static uint16_t calc_tcpip_csum(const void *ctx, uint8_t *packet, int len)
+static uint16_t calc_tcpip_csum(const void *ctx, const uint8_t *packet, int len)
 {
 	uint8_t *buf;
 	uint16_t csum;
@@ -84,7 +87,7 @@ static uint16_t calc_tcpip_csum(const void *ctx, uint8_t *packet, int len)
 }
 
 /* Show some ip packet details */
-static void debug_ip_packet(uint8_t *data, int len, int dir, char *info)
+static void debug_ip_packet(const uint8_t *data, int len, int dir, const char *info)
 {
 	uint8_t tcp_flags;
 	char flags_debugmsg[256];
@@ -221,7 +224,8 @@ struct defrag_queue_entry {
 LLIST_HEAD(gprs_sndcp_entities);
 
 /* Check if any compression parameters are set in the sgsn configuration */
-static inline int any_pcomp_or_dcomp_active(struct sgsn_instance *sgsn) {
+static inline int any_pcomp_or_dcomp_active(const struct sgsn_instance *sgsn)
+{
 	if (sgsn->cfg.pcomp_rfc1144.active || sgsn->cfg.pcomp_rfc1144.passive ||
 	    sgsn->cfg.dcomp_v42bis.active || sgsn->cfg.dcomp_v42bis.passive)
 		return true;
@@ -260,7 +264,7 @@ static int defrag_enqueue(struct gprs_sndcp_entity *sne, uint8_t seg_nr,
 }
 
 /* return if we have all segments of this N-PDU */
-static int defrag_have_all_segments(struct gprs_sndcp_entity *sne)
+static int defrag_have_all_segments(const struct gprs_sndcp_entity *sne)
 {
 	uint32_t seg_needed = 0;
 	unsigned int i;
@@ -275,7 +279,7 @@ static int defrag_have_all_segments(struct gprs_sndcp_entity *sne)
 	return 0;
 }
 
-static struct defrag_queue_entry *defrag_get_seg(struct gprs_sndcp_entity *sne,
+static struct defrag_queue_entry *defrag_get_seg(const struct gprs_sndcp_entity *sne,
 						 uint32_t seg_nr)
 {
 	struct defrag_queue_entry *dqe;
@@ -289,13 +293,62 @@ static struct defrag_queue_entry *defrag_get_seg(struct gprs_sndcp_entity *sne,
 	return NULL;
 }
 
+/* Returns talloced buffer containing decompressed data, NULL on error. */
+static uint8_t *decompress_segment(struct gprs_sndcp_entity *sne, void *ctx,
+				   const uint8_t *compressed_data, unsigned int compressed_data_len,
+				   unsigned int *decompressed_data_len)
+{
+	int rc;
+	uint8_t *expnd = NULL;
+	*decompressed_data_len = 0;
+
+#if DEBUG_IP_PACKETS == 1
+	DEBUGP(DSNDCP, "\n");
+	DEBUGP(DSNDCP, ":::::::::::::::::::::::::::::::::::::::::::::::::::\n");
+	DEBUGP(DSNDCP, "===================================================\n");
+#endif
+
+	expnd = talloc_zero_size(ctx, compressed_data_len * MAX_DATADECOMPR_FAC +
+					 MAX_HDRDECOMPR_INCR);
+	memcpy(expnd, compressed_data, compressed_data_len);
+
+	/* Apply data decompression */
+	rc = gprs_sndcp_dcomp_expand(expnd, compressed_data_len, sne->defrag.dcomp,
+				     sne->defrag.data);
+	if (rc < 0) {
+		LOGP(DSNDCP, LOGL_ERROR,
+			"Data decompression failed!\n");
+		talloc_free(expnd);
+		return NULL;
+	}
+
+	/* Apply header decompression */
+	rc = gprs_sndcp_pcomp_expand(expnd, rc, sne->defrag.pcomp, sne->defrag.proto);
+	if (rc < 0) {
+		LOGP(DSNDCP, LOGL_ERROR,
+			"TCP/IP Header decompression failed!\n");
+		talloc_free(expnd);
+		return NULL;
+	}
+
+	*decompressed_data_len = rc;
+
+#if DEBUG_IP_PACKETS == 1
+	debug_ip_packet(expnd, *decompressed_data_len, 1, "defrag_segments()");
+	DEBUGP(DSNDCP, "===================================================\n");
+	DEBUGP(DSNDCP, ":::::::::::::::::::::::::::::::::::::::::::::::::::\n");
+	DEBUGP(DSNDCP, "\n");
+#endif
+	return expnd;
+}
+
 /* Perform actual defragmentation and create an output packet */
 static int defrag_segments(struct gprs_sndcp_entity *sne)
 {
 	struct msgb *msg;
 	unsigned int seg_nr;
 	uint8_t *npdu;
-	int npdu_len;
+	unsigned int npdu_len;
 	int rc;
 	uint8_t *expnd = NULL;
 
@@ -336,53 +389,20 @@ static int defrag_segments(struct gprs_sndcp_entity *sne)
 	 * hands it off to the correct GTP tunnel + GGSN via gtp_data_req() */
 
 	/* Decompress packet */
-#if DEBUG_IP_PACKETS == 1
-	DEBUGP(DSNDCP, "                                                   \n");
-	DEBUGP(DSNDCP, ":::::::::::::::::::::::::::::::::::::::::::::::::::\n");
-	DEBUGP(DSNDCP, "===================================================\n");
-#endif
 	if (any_pcomp_or_dcomp_active(sgsn)) {
-
-		expnd = talloc_zero_size(msg, npdu_len * MAX_DATADECOMPR_FAC +
-					 MAX_HDRDECOMPR_INCR);
-		memcpy(expnd, npdu, npdu_len);
-
-		/* Apply data decompression */
-		rc = gprs_sndcp_dcomp_expand(expnd, npdu_len, sne->defrag.dcomp,
-					     sne->defrag.data);
-		if (rc < 0) {
-			LOGP(DSNDCP, LOGL_ERROR,
-			     "Data decompression failed!\n");
-			talloc_free(expnd);
-			return -EIO;
+		expnd = decompress_segment(sne, msg, npdu, npdu_len, &npdu_len);
+		if (!expnd) {
+			rc = -EIO;
+			goto ret_free;
 		}
-
-		/* Apply header decompression */
-		rc = gprs_sndcp_pcomp_expand(expnd, rc, sne->defrag.pcomp,
-					     sne->defrag.proto);
-		if (rc < 0) {
-			LOGP(DSNDCP, LOGL_ERROR,
-			     "TCP/IP Header decompression failed!\n");
-			talloc_free(expnd);
-			return -EIO;
-		}
-
-		/* Modify npu length, expnd is handed directly handed
-		 * over to gsn_rx_sndcp_ud_ind(), see below */
-		npdu_len = rc;
-	} else
+	} else {
 		expnd = npdu;
-#if DEBUG_IP_PACKETS == 1
-	debug_ip_packet(expnd, npdu_len, 1, "defrag_segments()");
-	DEBUGP(DSNDCP, "===================================================\n");
-	DEBUGP(DSNDCP, ":::::::::::::::::::::::::::::::::::::::::::::::::::\n");
-	DEBUGP(DSNDCP, "                                                   \n");
-#endif
+	}
 
-	/* Hand off packet to gtp */
-	rc = sgsn_rx_sndcp_ud_ind(&sne->ra_id, sne->lle->llme->tlli,
-				  sne->nsapi, msg, npdu_len, expnd);
+	/* Hand off packet to SGSN (SNDCP SN-UNITDATA.ind), which will forward it to GGSN (GTP): */
+	rc = sndcp_sn_unitdata_ind(sne, msg, npdu_len, expnd);
 
+ret_free:
 	/* we must free the memory we allocated above; ownership is not transferred
 	 * downwards in the call above */
 	msgb_free(msg);
@@ -522,7 +542,7 @@ int sndcp_sm_activate_ind(struct gprs_llc_lle *lle, uint8_t nsapi)
 }
 
 /* Entry point for the SNSM-DEACTIVATE.indication */
-int sndcp_sm_deactivate_ind(struct gprs_llc_lle *lle, uint8_t nsapi)
+int sndcp_sm_deactivate_ind(const struct gprs_llc_lle *lle, uint8_t nsapi)
 {
 	struct gprs_sndcp_entity *sne;
 
@@ -545,7 +565,7 @@ int sndcp_sm_deactivate_ind(struct gprs_llc_lle *lle, uint8_t nsapi)
 }
 
 /* Clean up all gprs_sndcp_entities related to llme (OS#4824) */
-void gprs_sndcp_sm_deactivate_ind_by_llme(struct gprs_llc_llme *llme)
+void gprs_sndcp_sm_deactivate_ind_by_llme(const struct gprs_llc_llme *llme)
 {
 	struct gprs_sndcp_entity *sne, *sne2;
 
@@ -665,7 +685,7 @@ static int sndcp_send_ud_frag(struct sndcp_frag_state *fs,
 }
 
 /* Request transmission of a SN-PDU over specified LLC Entity + SAPI */
-int sndcp_unitdata_req(struct msgb *msg, struct gprs_llc_lle *lle, uint8_t nsapi,
+int sndcp_sn_unitdata_req(struct msgb *msg, struct gprs_llc_lle *lle, uint8_t nsapi,
 			void *mmcontext)
 {
 	struct gprs_sndcp_entity *sne;
@@ -723,13 +743,14 @@ int sndcp_unitdata_req(struct msgb *msg, struct gprs_llc_lle *lle, uint8_t nsapi
 
 	sne = gprs_sndcp_entity_by_lle(lle, nsapi);
 	if (!sne) {
-		LOGP(DSNDCP, LOGL_ERROR, "Cannot find SNDCP Entity\n");
+		LOGP(DSNDCP, LOGL_ERROR, "Cannot find SNDCP Entity (lle=%p, TLLI=%08x, SAPI=%u, NSAPI=%u)\n",
+			 lle, lle->llme->tlli, lle->sapi, nsapi);
 		msgb_free(msg);
 		return -EIO;
 	}
 
 	/* Check if we need to fragment this N-PDU into multiple SN-PDUs */
-	if (msg->len > lle->params.n201_u - 
+	if (msg->len > lle->params.n201_u -
 			(sizeof(*sch) + sizeof(*suh) + sizeof(*scomph))) {
 		/* initialize the fragmenter state */
 		fs.msg = msg;
@@ -774,7 +795,7 @@ int sndcp_unitdata_req(struct msgb *msg, struct gprs_llc_lle *lle, uint8_t nsapi
 }
 
 /* Section 5.1.2.17 LL-UNITDATA.ind */
-int sndcp_llunitdata_ind(struct msgb *msg, struct gprs_llc_lle *lle,
+int sndcp_ll_unitdata_ind(struct msgb *msg, struct gprs_llc_lle *lle,
 			 uint8_t *hdr, uint16_t len)
 {
 	struct gprs_sndcp_entity *sne;
@@ -844,63 +865,40 @@ int sndcp_llunitdata_ind(struct msgb *msg, struct gprs_llc_lle *lle,
 		LOGP(DSNDCP, LOGL_ERROR, "Short SNDCP N-PDU: %d\n", npdu_len);
 		return -EIO;
 	}
-	/* actually send the N-PDU to the SGSN core code, which then
-	 * hands it off to the correct GTP tunnel + GGSN via gtp_data_req() */
+	/* actually send the N-PDU to the SGSN core code (SNDCP SN-UNITDATA.ind) */
 
 	/* Decompress packet */
-#if DEBUG_IP_PACKETS == 1
-	DEBUGP(DSNDCP, "                                                   \n");
-	DEBUGP(DSNDCP, ":::::::::::::::::::::::::::::::::::::::::::::::::::\n");
-	DEBUGP(DSNDCP, "===================================================\n");
-#endif
 	if (any_pcomp_or_dcomp_active(sgsn)) {
-
-		expnd = talloc_zero_size(msg, npdu_len * MAX_DATADECOMPR_FAC +
-					 MAX_HDRDECOMPR_INCR);
-		memcpy(expnd, npdu, npdu_len);
-
-		/* Apply data decompression */
-		rc = gprs_sndcp_dcomp_expand(expnd, npdu_len, sne->defrag.dcomp,
-					     sne->defrag.data);
-		if (rc < 0) {
-			LOGP(DSNDCP, LOGL_ERROR,
-			     "Data decompression failed!\n");
-			talloc_free(expnd);
-			return -EIO;
+		expnd = decompress_segment(sne, msg, npdu, npdu_len, (unsigned int *)&npdu_len);
+		if (!expnd) {
+			rc = -EIO;
+			goto ret_free;
 		}
-
-		/* Apply header decompression */
-		rc = gprs_sndcp_pcomp_expand(expnd, rc, sne->defrag.pcomp,
-					     sne->defrag.proto);
-		if (rc < 0) {
-			LOGP(DSNDCP, LOGL_ERROR,
-			     "TCP/IP Header decompression failed!\n");
-			talloc_free(expnd);
-			return -EIO;
-		}
-
-		/* Modify npu length, expnd is handed directly handed
-		 * over to gsn_rx_sndcp_ud_ind(), see below */
-		npdu_len = rc;
-	} else
+	} else {
 		expnd = npdu;
-#if DEBUG_IP_PACKETS == 1
-	debug_ip_packet(expnd, npdu_len, 1, "sndcp_llunitdata_ind()");
-	DEBUGP(DSNDCP, "===================================================\n");
-	DEBUGP(DSNDCP, ":::::::::::::::::::::::::::::::::::::::::::::::::::\n");
-	DEBUGP(DSNDCP, "                                                   \n");
-#endif
+	}
 
 	/* Hand off packet to gtp */
-	rc = sgsn_rx_sndcp_ud_ind(&sne->ra_id, lle->llme->tlli,
-				  sne->nsapi, msg, npdu_len, expnd);
+	rc = sndcp_sn_unitdata_ind(sne, msg, npdu_len, expnd);
 
+ret_free:
 	if (any_pcomp_or_dcomp_active(sgsn))
 		talloc_free(expnd);
 
 	return rc;
 }
 
+/* 5.1.1.4 SN-UNITDATA.indication
+ * Called by SNDCP when it has received/re-assembled a N-PDU
+ */
+int sndcp_sn_unitdata_ind(struct gprs_sndcp_entity *sne,
+		    struct msgb *msg, uint32_t npdu_len, uint8_t *npdu)
+{
+	/* Hand it off N-PDU to the correct GTP tunnel + GGSN: */
+	return sgsn_gtp_data_req(&sne->ra_id, sne->lle->llme->tlli,
+				  sne->nsapi, msg, npdu_len, npdu);
+}
+
 #if 0
 /* Section 5.1.2.1 LL-RESET.ind */
 static int sndcp_ll_reset_ind(struct gprs_sndcp_entity *se)
@@ -1056,7 +1054,7 @@ int sndcp_sn_xid_req(struct gprs_llc_lle *lle, uint8_t nsapi)
 
 /* Handle header compression entites */
 static int handle_pcomp_entities(struct gprs_sndcp_comp_field *comp_field,
-				 struct gprs_llc_lle *lle)
+				 const struct gprs_llc_lle *lle)
 {
 	/* Note: This functions also transforms the comp_field into its
 	 * echo form (strips comp values, resets propose bit etc...)
@@ -1106,7 +1104,7 @@ static int handle_pcomp_entities(struct gprs_sndcp_comp_field *comp_field,
 
 /* Hanle data compression entites */
 static int handle_dcomp_entities(struct gprs_sndcp_comp_field *comp_field,
-				 struct gprs_llc_lle *lle)
+				 const struct gprs_llc_lle *lle)
 {
 	/* See note in handle_pcomp_entities() */
 
@@ -1148,7 +1146,7 @@ static int handle_dcomp_entities(struct gprs_sndcp_comp_field *comp_field,
  * (See also: TS 144 065, Section 6.8 XID parameter negotiation) */
 int sndcp_sn_xid_ind(struct gprs_llc_xid_field *xid_field_indication,
 		     struct gprs_llc_xid_field *xid_field_response,
-		     struct gprs_llc_lle *lle)
+		     const struct gprs_llc_lle *lle)
 {
 	/* Note: This function computes the SNDCP-XID response that is sent
 	 * back to the ms when a ms originated XID is received. The
