@@ -30,7 +30,7 @@
 #include <osmocom/gsupclient/gsup_client.h>
 
 #include <osmocom/sgsn/sgsn.h>
-#include <osmocom/sgsn/gprs_sgsn.h>
+#include <osmocom/sgsn/mmctx.h>
 #include <osmocom/sgsn/gprs_gmm.h>
 #include <osmocom/sgsn/gprs_utils.h>
 
@@ -48,8 +48,6 @@
 	     (gsup)->imsi, \
 	     ## args)
 
-extern void *tall_sgsn_ctx;
-
 LLIST_HEAD(_gprs_subscribers);
 struct llist_head * const gprs_subscribers = &_gprs_subscribers;
 
